from collections import UserDict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import ruamel.yaml  # type: ignore


@dataclass
class PipelineUnit:
    type: str
    args: dict[str, Any] = field(default_factory=dict)


@dataclass
class DataConfig:
    label_column: str = "label"
    text_column: str | list[str] = "text"


@dataclass
class BunruijaConfig:
    pipeline: list[PipelineUnit]
    output_dir: Path
    data: DataConfig | None = None
    dataset_args: UserDict | None = None

    @classmethod
    def from_yaml(cls, config_file):
        with open(config_file) as f:
            yaml = ruamel.yaml.YAML()
            config = yaml.load(f)

            label_column: str = config["data"].pop("label_column", "label")
            text_column: str | list[str] = config["data"].pop("text_column", "text")

            return cls(
                data=DataConfig(label_column=label_column, text_column=text_column),
                pipeline=[PipelineUnit(**unit) for unit in config["pipeline"]],
                output_dir=Path(config.get("output_dir", "output")),
                dataset_args=UserDict(config["data"]["args"]),
            )



from setuptools import Extension  # type: ignore
from distutils.command.build_ext import build_ext

from Cython.Build import cythonize  # type: ignore


def build(setup_kwargs):
    ext_modules = cythonize(
        [
            Extension(
                "bunruija.modules.vector_processor",
                sources=(
                    ["bunruija/modules/vector_processor.pyx"]
                    + ["csrc/keyed_vector.cc", "csrc/string_util.cc"]
                ),
                libraries=["sqlite3"],
                include_dirs=["include"],
                extra_compile_args=["-std=c++11"],
                extra_link_args=["-std=c++11"],
            ),
        ]
    )

    setup_kwargs.update(
        {"ext_modules": ext_modules, "cmdclass": {"build_ext": build_ext}}
    )



class BaseTokenizer:
    def __init__(self, name: str | None = None):
        if name:
            self.name = name

    @classmethod
    def module_name(cls) -> str:
        module_name = ".".join([cls.__module__, cls.__name__])
        return module_name

    def __call__(self, text: str) -> list[str]:
        raise NotImplementedError



class Dictionary:
    def __init__(self, pad="<pad>", eos="</s>", bos="<s>"):
        self.elements = []
        self.count = []
        self.index_to_element = {}

        self.pad = pad
        self.eos = eos
        self.bos = bos

        self.add(self.pad)
        self.add(self.eos)
        self.add(self.bos)

    def __contains__(self, element):
        return element in self.index_to_element

    def add(self, element, n=1):
        if element not in self:
            self.elements.append(element)
            index = len(self.index_to_element)
            self.index_to_element[element] = index
            self.count.append(n)
        else:
            index = self.get_index(element)
            self.count[index] += n
        return index

    def get_element(self, index):
        return self.elements[index]

    def get_index(self, element):
        return self.index_to_element[element]

    def __len__(self):
        return len(self.elements)



# distutils: language = c++
from functools import lru_cache

import cython
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.unordered_map cimport unordered_map
from libcpp.vector cimport vector

import numpy as np


cdef extern from "bunruija/status.h" namespace "bunruija":
    cdef cppclass CppStatus "bunruija::Status":
        CppStatus() except +
        CppStatus(int, string) except +
        int status_code
        string status_message

cdef extern from "bunruija/keyed_vector.h" namespace "bunruija":
    cdef cppclass CppPretrainedVectorProcessor "bunruija::PretrainedVectorProcessor":
        CppPretrainedVectorProcessor() except +
        CppStatus convert(char *, char *)
        CppStatus load(string)
        CppStatus query(string, vector[float] *)
        CppStatus batch_query(vector[string], unordered_map[string, vector[float]] *)
        CppStatus contains(string, bool *)
        int dim()
        int length()


cdef class Status:
    cdef CppStatus *thisptr

    def __cinit__(self, code=0, message=""):
        cdef string message_ = string(bytes(message))
        self.thisptr = new CppStatus(code, message_)

    def __repr__(self):
        msg = self.thisptr.status_message.decode('utf-8')
        return f'Status(status_code={self.thisptr.status_code}, status_message=\"{msg}\")'


@cython.auto_pickle(True)
cdef class PretrainedVectorProcessor:
    cdef CppPretrainedVectorProcessor *thisptr

    def __init__(self):
        self.thisptr = new CppPretrainedVectorProcessor()

    def convert(self, db_file, input_file):
        status_ = self.thisptr.convert(db_file.encode('utf-8'), input_file.encode('utf-8'))
        status = Status(status_.status_code, status_.status_message)
        return status

    def load(self, input_file):
        cdef string input_file_ = string(bytes(input_file.encode('utf-8')))
        status_ = self.thisptr.load(input_file_)
        status = Status(status_.status_code, status_.status_message)
        return status

    @lru_cache(maxsize=0)
    def query(self, word):
        cdef string word_ = string(bytes(word.encode('utf-8')))
        cdef vector[float] vec_;
        status_ = self.thisptr.query(word_, &vec_)
        status = Status(status_.status_code, status_.status_message)

        vec = np.asarray(vec_)
        return vec, status

    @lru_cache(maxsize=0)
    def batch_query(self, words):
        cdef unordered_map[string, vector[float]] vec_;
        status_ = self.thisptr.batch_query([bytes(w.encode("utf-8")) for w in words], &vec_)
        status = Status(status_.status_code, status_.status_message)

        vecs = []
        for word in words:
            v = vec_[bytes(word.encode("utf-8"))]
            if len(v) == 0:
                v = [0 for _ in range(self.emb_dim)]
            vecs.append(v)
#             print(word, "shape", len(v), type(v), type(v[0]))
        vecs = np.asarray(vecs)
#         print("shape", vecs.shape)

        return vecs, status

    def __contains__(self, word):
        cdef string word_ = string(bytes(word.encode('utf-8')))
        cdef bool out_;
        status_ = self.thisptr.contains(word_, &out_)
        status = Status(status_.status_code, status_.status_message)
        return out_

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)

    def __reduce_ex__(self, proto):
        func = PretrainedVectorProcessor
        args = ()
        state = self.__getstate__()
        listitems = None
        dictitems = None
        return (func, args, state, listitems, dictitems)

    @property
    def emb_dim(self):
        return self.thisptr.dim()

    @property
    def length(self):
        return self.thisptr.length()



import logging
import sys

from . import Evaluator, options

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
    stream=sys.stdout,
)
logger = logging.getLogger("bunruija.evaluate")


def main(args):
    parser = options.get_default_evaluation_parser()
    args = parser.parse_args(args)

    evaluator = Evaluator(args)
    evaluator.evaluate()


def cli_main():
    main(sys.argv[1:])



import random

import mmh3  # type: ignore
import torch

from ..classifier import NeuralBaseClassifier  # type: ignore


class Hasher:
    def __init__(self, n_features=64):
        self.n_features = n_features

    def get_hash_codes(self, word):
        hash_ = mmh3.hash128(word)
        return hash_


class WeightMask:
    def __init__(self, index):
        self.index = index

    def __call__(self, module, _):
        mask = module.raw_weight.new_ones(module.raw_weight.size())
        if self.index.device != mask.device:
            mask.index_fill_(2, self.index.to(mask.device), 0.0)
        else:
            mask.index_fill_(2, self.index, 0.0)
        module.weight = module.raw_weight * mask


class ConvolutionLayer(torch.nn.Module):
    def __init__(self, kernel_size, dim_hid):
        super().__init__()
        self.dim_hid = dim_hid
        self.conv = torch.nn.Conv2d(
            1, self.dim_hid, kernel_size=(kernel_size, self.dim_hid), stride=1
        )

        self.batch_norm = torch.nn.BatchNorm1d(self.dim_hid)

    def forward(self, x):
        x = self.conv(x).squeeze(3)
        x = self.batch_norm(x)
        return x


class ProjectAttentionLayer(torch.nn.Module):
    def __init__(self, kernel_size, dim_hid, skip_bigram=None):
        super().__init__()
        self.conv_value = ConvolutionLayer(kernel_size, dim_hid)

        if isinstance(skip_bigram, list):
            self.conv_value.conv.raw_weight = self.conv_value.conv.weight
            del self.conv_value.conv.weight
            weight_mask = WeightMask(torch.tensor(skip_bigram))
            self.conv_value.conv.register_forward_pre_hook(weight_mask)
        elif skip_bigram is None:
            pass
        else:
            raise ValueError(skip_bigram)

        self.conv_attn = ConvolutionLayer(kernel_size, dim_hid)
        self.zero_pad = torch.nn.ZeroPad2d((0, 0, kernel_size - 1, 0))

    def forward(self, x_value_in, x_attn_in):
        # (bsz, output_channel, seq_len)
        x_value_in = self.zero_pad(x_value_in)
        x_value = self.conv_value(x_value_in)

        # (bsz, output_channel, seq_len)
        x_attn_in = self.zero_pad(x_attn_in)
        x_attn = self.conv_attn(x_attn_in)
        x_attn = torch.softmax(x_attn, dim=2)

        x = x_attn * x_value
        # (bsz, output_channel)
        x = torch.sum(x, dim=2)
        return x


class ProjectedAttention(torch.nn.Module):
    def __init__(self, kernel_sizes, dim_hid, skip_bigrams=None):
        super().__init__()
        self.dim_hid = dim_hid

        if isinstance(skip_bigrams, list):
            self.layers = torch.nn.ModuleList(
                [
                    ProjectAttentionLayer(kernel_size, dim_hid, skip_bigram=skip_bigram)
                    for kernel_size, skip_bigram in zip(kernel_sizes, skip_bigrams)
                ]
            )
        else:
            self.layers = torch.nn.ModuleList(
                [
                    ProjectAttentionLayer(kernel_size, dim_hid, skip_bigram=None)
                    for kernel_size in kernel_sizes
                ]
            )

    def forward(self, x_value_in, x_attn_in):
        x_list = []

        for layer in self.layers:
            x = layer(x_value_in, x_attn_in)
            x_list.append(x)
        x = torch.cat(x_list, dim=1)
        return x


class StringProjector(torch.nn.Module):
    def __init__(self, n_features, distortion_probability=0.25):
        super().__init__()
        self.n_features = n_features
        self.hasher = Hasher(self.n_features)
        self.distort = distortion_probability
        self.mapping_table = [0, 1, -1, 0]
        self.random_char = None

    def word_string_distort(self, word):
        if self.distort == 0 or len(word) == 0:
            return word
        else:
            if random.random() < self.distort:
                distortion_type = random.random()
                rindex = random.randint(0, len(word) - 1)
                if distortion_type < 0.33:
                    self.random_char = word[rindex]
                    word = word[:rindex] + word[rindex + 1 :]
                elif distortion_type < 0.66:
                    if len(word) > 2:
                        self.random_char = word[rindex]
                        rindex2 = random.randint(0, len(word) - 1)
                        word = list(word)
                        word[rindex2] = word[rindex]
                        word = "".join(word)
                elif self.random_char:
                    word = list(word)
                    word[rindex] = self.random_char
                    word = "".join(word)
            return word

    def forward(self, batch_words):
        max_seq_len = max(len(words) for words in batch_words)
        x = torch.zeros(
            len(batch_words), max_seq_len, self.n_features, dtype=torch.float32
        )

        for batch_idx, words in enumerate(batch_words):
            for t, word in enumerate(words):
                if self.training:
                    word = self.word_string_distort(word)

                hash_ = self.hasher.get_hash_codes(word)
                projection = []

                b = self.mapping_table[hash_ & 0x3]
                projection.append(b)

                for i in range(1, self.n_features):
                    hash_ >>= 2
                    b = self.mapping_table[hash_ & 0x3]
                    projection.append(b)
                x[batch_idx, t] = torch.tensor(projection)
        return x


class PRADO(NeuralBaseClassifier):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.random_char = None

        self.n_features = kwargs.get("n_features", 512)
        distort = kwargs.get("distortion_probability", 0.25)

        self.string_proj = StringProjector(self.n_features, distort)

        self.dim_emb = kwargs.get("dim_emb", 64)
        self.dim_hid = kwargs.get("dim_hid", 64)

        self.fc_value = torch.nn.Linear(self.n_features, self.dim_hid)
        self.fc_attn = torch.nn.Linear(self.n_features, self.dim_hid)

        dropout = kwargs.get("dropout", 0.15)
        self.dropout = torch.nn.Dropout(p=dropout)

        self.kernel_sizes = kwargs.get("kernel_sizes", [2, 3, 3, 4])
        self.skip_bigrams = kwargs.get("skip_bigrams", [None, None, [1], [1, 2]])
        self.attention = ProjectedAttention(
            self.kernel_sizes, self.dim_hid, skip_bigrams=self.skip_bigrams
        )

        self.batch_norm_value = torch.nn.BatchNorm1d(self.dim_hid)
        self.batch_norm_attn = torch.nn.BatchNorm1d(self.dim_hid)

    def train(self, mode=True):
        super().train(mode)

    def init_layer(self, data):
        self.pad = 0
        self.fc = torch.nn.Linear(
            len(self.kernel_sizes) * self.dim_hid, len(self.labels), bias=True
        )

    def forward(self, batch):
        projection = self.string_proj(batch["words"])

        projection = projection.to(batch["inputs"].device)
        mask = (batch["inputs"] == self.pad).unsqueeze(2)

        x_value = self.fc_value(projection)
        x_value = self.dropout(x_value)
        x_value = x_value.masked_fill_(mask, 0)
        x_value = self.batch_norm_value(x_value.transpose(1, 2)).transpose(1, 2)
        x_value = torch.relu(x_value)

        x_attn = self.fc_attn(projection)
        x_attn = self.dropout(x_attn)
        x_attn = x_attn.masked_fill_(mask, 0)
        x_attn = self.batch_norm_attn(x_attn.transpose(1, 2)).transpose(1, 2)
        x_attn = torch.relu(x_attn)

        x_value = x_value.unsqueeze(1)
        x_attn = x_attn.unsqueeze(1)

        x = self.attention(x_value, x_attn)
        x = self.fc(x)
        return x



from .filter import BaseFilter


class PosFilter(BaseFilter):
    def __init__(self, exclude_pos=[]):
        super().__init__(name="pos")
        self.exclude_pos = exclude_pos

    def __repr__(self):
        if len(self.exclude_pos) > 0:
            args = f'exclude_pos=[{", ".join(self.exclude_pos)}]'
        else:
            args = ""
        out = f"{self.__class__.__name__}({args})"
        return out

    def __call__(self, surface, features, *args, **kwargs):
        for p in self.exclude_pos:
            if features[0] == p:
                return True
        return False



from .tokenizer import BaseTokenizer


class SpaceTokenizer(BaseTokenizer):
    def __init__(self, reduce_redundant_spaces: bool = True, **kwargs):
        self.reduce_redundant_spaces = reduce_redundant_spaces
        super().__init__(name="space")

    def __call__(self, text) -> list[str]:
        if self.reduce_redundant_spaces:
            result = [token for token in text.split(" ") if token != ""]
            return result
        else:
            return text.split(" ")

    def __repr__(self) -> str:
        out = f"{self.__class__.__name__}()"
        return out



from typing import Any, Callable

import numpy as np
import transformers  # type: ignore
from scipy.sparse import csr_matrix  # type: ignore
from sklearn.base import TransformerMixin  # type: ignore

from ..data import Dictionary
from ..tokenizers import MeCabTokenizer


class SequenceVectorizer(TransformerMixin):
    def __init__(
        self,
        tokenizer: Callable[[str], list[str]] | None = None,
        max_features: int | None = None,
        keep_raw_word: bool = True,
        only_raw_word: bool = False,
        dictionary: Dictionary = Dictionary(),
        **kwargs,
    ):
        super().__init__()

        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.vocabulary_ = dictionary.index_to_element
        self.max_features = max_features
        self.keep_raw_word = keep_raw_word
        self.only_raw_word = only_raw_word

    @classmethod
    def module_name(cls) -> str:
        module_name = ".".join([cls.__module__, cls.__name__])
        return module_name

    def __repr__(self) -> str:
        args = []
        if self.tokenizer:
            args.append(f"tokenizer={self.tokenizer}")
        if self.max_features:
            args.append(f"max_features={self.max_features}")
        args.append(f"keep_raw_word={self.keep_raw_word}")
        args.append(f"only_raw_word={self.only_raw_word}")
        out = f'{self.__class__.__name__}({", ".join(args)})'
        return out

    def build_tokenizer(self) -> Callable[[str], list[str]]:
        if self.tokenizer is not None:
            return self.tokenizer

        self.tokenizer = MeCabTokenizer()
        return self.tokenizer

    def set_params(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def get_params(self, deep=True) -> dict[str, Any]:
        return {
            "tokenizer": self.tokenizer,
            "max_features": self.max_features,
            "dictionary": self.dictionary,
            "keep_raw_word": self.keep_raw_word,
            "only_raw_word": self.only_raw_word,
        }

    def fit(self, raw_documents: list[str], y=None) -> "SequenceVectorizer":
        if self.only_raw_word:
            return self

        tokenizer = self.build_tokenizer()

        for row_id, document in enumerate(raw_documents):
            elements = tokenizer(document)
            for element in elements:
                self.dictionary.add(element)

        if self.max_features is not None:
            filtered_dict = Dictionary()
            for k, v in sorted(
                zip(self.dictionary.elements, self.dictionary.count),
                key=lambda x: x[1],
                reverse=True,
            )[: self.max_features]:
                filtered_dict.add(k, v)

            self.dictionary = filtered_dict
        return self

    def transform(
        self,
        raw_documents: list[str],
    ) -> csr_matrix | tuple[csr_matrix, list[str]]:
        data = []
        raw_words = []
        row = []
        col = []
        max_col = 0

        tokenizer = self.build_tokenizer()
        for row_id, document in enumerate(raw_documents):
            if isinstance(tokenizer, transformers.PreTrainedTokenizerBase):
                elements = tokenizer(document, truncation=True)
            else:
                elements = tokenizer(document)

            if isinstance(elements, transformers.tokenization_utils_base.BatchEncoding):
                input_ids = elements["input_ids"]
                max_col = max(max_col, len(input_ids))

                for i, index in enumerate(input_ids):
                    data.append(index)
                    row.append(row_id)
                    col.append(i)

            else:
                max_col = max(max_col, len(elements))

                for i, element in enumerate(elements):
                    if self.only_raw_word:
                        raw_words.append(element)
                        index = 1
                        data.append(index)
                        row.append(row_id)
                        col.append(i)
                    else:
                        if element in self.dictionary:
                            if self.keep_raw_word:
                                raw_words.append(element)

                            index = self.dictionary.get_index(element)

                            data.append(index)
                            row.append(row_id)
                            col.append(i)

        data = np.array(data)
        row = np.array(row)
        col = np.array(col)

        s = csr_matrix((data, (row, col)), shape=(len(raw_documents), max_col))
        if self.keep_raw_word:
            return s, raw_words
        else:
            return s

    def __call__(self, text):
        tokenizer = self.build_tokenizer()
        ret = tokenizer(text)
        return ret



import sys
from logging import INFO, basicConfig, getLogger

from . import options
from .trainer import Trainer

basicConfig(
    format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=INFO,
    stream=sys.stdout,
)
logger = getLogger("bunruija.train")


def main(args):
    parser = options.get_default_train_parser()
    args = parser.parse_args(args)

    trainer = Trainer(args.yaml)
    trainer.train()


def cli_main():
    main(sys.argv[1:])


if __name__ == "__main__":
    cli_main()



import logging
import os

import torch

from bunruija.modules.vector_processor import PretrainedVectorProcessor  # type: ignore

logger = logging.getLogger(__name__)


class StaticEmbedding(torch.nn.Module):
    def __init__(self, embedding_path):
        super().__init__()

        self.processor = PretrainedVectorProcessor()

        self.embedding_path = embedding_path
        self.convert(self.embedding_path)
        self.processor.load(self.embedding_path + ".bunruija")
        self.dim_emb = self.processor.emb_dim
        self.length = self.processor.length

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)
        self.processor.load(self.embedding_path + ".bunruija")

    def convert(self, embedding_path):
        ext = ".bunruija"
        output_file = embedding_path + ext

        if not os.path.exists(output_file):
            logger.info(f"Converting word vectors to {output_file}")
            logger.info("This conversion will take a while")
            self.processor.convert(output_file, embedding_path)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(vocab_size={self.length}, embedding_size={self.dim_emb})"
        )

    def forward(self, batch):
        bsz = len(batch["words"])
        max_seq_len = 0
        for batch_idx in range(len(batch["words"])):
            max_seq_len = max(max_seq_len, len(batch["words"][batch_idx]))

        embed = torch.zeros((bsz, max_seq_len, self.dim_emb))
        for batch_idx in range(len(batch["words"])):
            vec, status = self.processor.batch_query(tuple(batch["words"][batch_idx]))
            seq_len = len(batch["words"][batch_idx])
            embed[batch_idx, :seq_len] = torch.from_numpy(vec)

        return embed



from argparse import Namespace

import datasets
import numpy as np
from sklearn.metrics import classification_report, confusion_matrix

from .data.dataset import load_data
from .dataclass import BunruijaConfig
from .predictor import Predictor


class Evaluator:
    """Evaluates a trained model"""

    def __init__(self, args: Namespace):
        self.config = BunruijaConfig.from_yaml(args.yaml)
        self.verbose = args.verbose
        self.predictor = Predictor.from_pretrained(self.config.output_dir)

    def evaluate(self):
        labels_test, X_test = load_data(
            self.config.dataset_args,
            split=datasets.Split.TEST,
            label_column=self.config.data.label_column,
            text_column=self.config.data.text_column,
        )

        y_test: np.ndarray = self.predictor.label_encoder.transform(labels_test)
        y_pred: np.ndarray = self.predictor(X_test)

        labels = list(self.predictor.label_encoder.classes_)
        if self.verbose:
            conf_mat = confusion_matrix(y_test, y_pred)
            for i in range(len(labels)):
                print("True", "Pred", "Num samples", sep="\t")
                for j in range(len(labels)):
                    print(labels[i], labels[j], conf_mat[i, j], sep="\t")
                print()

        report: str = classification_report(
            y_test,
            y_pred,
            target_names=labels,
        )
        print(report)



import datasets
import numpy as np
import sklearn  # type: ignore
from sklearn.preprocessing import LabelEncoder  # type: ignore

from . import BunruijaConfig, PipelineBuilder, Saver
from .data.dataset import load_data


class Trainer:
    """Trains a text classification model."""

    def __init__(self, config_file: str):
        self.config = BunruijaConfig.from_yaml(config_file)

        self.model = PipelineBuilder(self.config).build()
        self.saver = Saver(self.config)

    def train(self):
        labels_train, X_train = load_data(
            self.config.dataset_args,
            split=datasets.Split.TRAIN,
            label_column=self.config.data.label_column,
            text_column=self.config.data.text_column,
        )

        label_encoder = LabelEncoder()
        y_train: np.ndarray = label_encoder.fit_transform(labels_train)

        self.model.fit(X_train, y_train)

        self.saver(self.model, label_encoder)

        labels_dev, X_dev = load_data(
            self.config.dataset_args,
            split=datasets.Split.VALIDATION,
            label_column=self.config.data.label_column,
            text_column=self.config.data.text_column,
        )

        y_dev: np.ndarray = label_encoder.transform(labels_dev)

        y_pred = self.model.predict(X_dev)

        fscore = sklearn.metrics.f1_score(y_dev, y_pred, average="micro")
        print(f"F-score on dev: {fscore}")
        target_names: list[str] = list(label_encoder.classes_)
        report = sklearn.metrics.classification_report(
            y_pred, y_dev, target_names=target_names
        )
        print(report)



import pickle

from sklearn.preprocessing import LabelEncoder  # type: ignore

from .dataclass import BunruijaConfig


class Saver:
    def __init__(self, config: BunruijaConfig):
        self.config = config

    def __call__(self, model, label_encoder: LabelEncoder):
        if not self.config.output_dir.exists():
            self.config.output_dir.mkdir(parents=True)

        with open(self.config.output_dir / "model.bunruija", "wb") as f:
            model_data = {
                "pipeline": model,
                "label_encoder": label_encoder,
            }
            pickle.dump(model_data, f)



import numpy as np
import torch

from bunruija.classifiers.classifier import NeuralBaseClassifier
from bunruija.classifiers.qrnn.qrnn_layer import QRNNLayer
from bunruija.modules import StaticEmbedding


class QRNN(NeuralBaseClassifier):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.embedding_path: str | None = kwargs.get("static_embedding_path", None)

        self.dim_emb: int = kwargs.get("dim_emb", 256)
        self.dim_hid: int = kwargs.get("dim_hid", 128)
        self.dropout_prob: float = kwargs.get("dropout", 0.15)
        self.window_size: int = kwargs.get("window_size", 3)
        self.bidirectional: bool = kwargs.get("bidirectional", True)

        if self.embedding_path:
            self.static_embed = StaticEmbedding(self.embedding_path)
        else:
            self.static_embed = None

        self.dropout = torch.nn.Dropout(self.dropout_prob)
        self.layers = torch.nn.ModuleList()
        num_layers = kwargs.get("num_layers", 2)
        for i in range(num_layers):
            if i == 0:
                input_size = (
                    self.dim_emb + self.static_embed.dim_emb
                    if self.static_embed
                    else self.dim
                )
            else:
                input_size = 2 * self.dim_hid if self.bidirectional else self.dim_hid

            self.layers.append(
                QRNNLayer(
                    input_size,
                    self.dim_hid,
                    window_size=self.window_size,
                    bidirectional=self.bidirectional,
                )
            )

    def init_layer(self, data):
        self.pad = 0
        max_input_idx = 0
        for data_i in data:
            max_input_idx = max(max_input_idx, np.max(data_i["inputs"]))

        self.embed = torch.nn.Embedding(
            max_input_idx + 1,
            self.dim_emb,
            padding_idx=0,
        )

        self.out = torch.nn.Linear(
            2 * self.dim_hid if self.bidirectional else self.dim_hid,
            len(self.labels),
            bias=True,
        )

    def forward(self, batch):
        src_tokens = batch["inputs"]

        x = self.embed(src_tokens)
        if self.static_embed is not None:
            x_static = self.static_embed(batch)
            x_static = x_static.to(x.device)
            x = torch.cat([x, x_static], dim=2)

        x = self.dropout(x)

        for i, layer in enumerate(self.layers):
            x = layer(x)
            if i < len(self.layers) - 1:
                x = self.dropout(x)

        x = torch.nn.functional.adaptive_max_pool2d(
            x, (1, 2 * self.dim_hid if self.bidirectional else self.dim_hid)
        )
        x = x.squeeze(1)
        x = self.out(x)
        return x



import time
from functools import partial
from logging import getLogger

import numpy as np
import torch
import torch.nn.functional as F
from sklearn.base import BaseEstimator, ClassifierMixin  # type: ignore

logger = getLogger(__name__)


def _addindent(s_, numSpaces):
    s = s_.split("\n")
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * " ") + line for line in s]
    s = "\n".join(s)
    s = first + "\n" + s
    return s


class BaseClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self):
        super().__init__()

    def fit(self, X, y):
        raise NotImplementedError

    def predict(self, X):
        raise NotImplementedError


class Collator:
    def __init__(self, padding_value):
        self.padding_value = padding_value

    def __call__(self, samples):
        inputs = torch.nn.utils.rnn.pad_sequence(
            [torch.tensor(sample["inputs"]) for sample in samples],
            batch_first=True,
            padding_value=self.padding_value,
        )

        attention_mask = inputs != self.padding_value
        batch = {"inputs": inputs, "attention_mask": attention_mask}
        if "label" in samples[0]:
            labels = torch.tensor([sample["label"] for sample in samples])
            batch["labels"] = labels

        if "raw_words" in samples[0]:
            words = [sample["raw_words"] for sample in samples]
            batch["words"] = words
        return batch


def move_to_cuda(batch):
    for k, v in batch.items():
        if isinstance(v, torch.Tensor):
            batch[k] = v.cuda()
    return batch


class NeuralBaseClassifier(BaseClassifier, torch.nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

        self.kwargs = kwargs
        self.device = kwargs.get("device", "cpu")
        self.max_epochs = kwargs.get("max_epochs", 3)
        self.batch_size = kwargs.get("batch_size", 20)
        self.num_workers = kwargs.get("num_workers", 1)

        self.log_interval = kwargs.get("log_interval", 100)
        self.optimizer_type = kwargs.get("optimizer", "adam")
        self.save_every_step = kwargs.get("save_every_step", -1)
        self.saver = kwargs.get("saver", None)
        self.labels = set()
        logger.info(f"device: {self.device}")

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split("\n")
        child_lines = []
        for key, module in self._modules.items():
            mod_str = repr(module)
            mod_str = _addindent(mod_str, 2)
            child_lines.append("(" + key + "): " + mod_str)
        lines = extra_lines + child_lines

        main_str = self._get_name() + "("
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += "\n  " + "\n  ".join(lines) + "\n"

        main_str += ")"
        return main_str

    def init_layer(self, data):
        raise NotImplementedError

    def convert_data(self, X, y=None):
        if y is not None:
            logger.info("Loading data")

        # if len(X) == 2 and isinstance(X[1], list):
        if isinstance(X, tuple):
            indices = X[0]
            raw_words = X[1]
            has_raw_words = True
        else:
            has_raw_words = False
            indices = X
            raw_words = None

        data = []
        for i in range(len(indices.indptr) - 1):
            start = indices.indptr[i]
            end = indices.indptr[i + 1]
            data_i = {
                "inputs": indices.data[start:end],
            }

            if y is not None:
                data_i["label"] = y[i]
                self.labels.add(y[i])

            if has_raw_words:
                data_i["raw_words"] = raw_words[start:end]
            data.append(data_i)
        return data

    def fit(self, X, y):
        data = self.convert_data(X, y)
        self.init_layer(data)

        optimizer = self.build_optimizer()
        start_at = time.perf_counter()

        self.to(self.device)
        self.train()

        logger.info(f"{self}")
        num_parameters = sum(p.numel() for p in self.parameters() if p.requires_grad)
        logger.info(f"Num params: {num_parameters}")
        step = 0
        loss_accum = 0
        n_samples_accum = 0
        start_at_accum = time.perf_counter()

        collator = Collator(self.pad)
        for epoch in range(self.max_epochs):
            for batch in torch.utils.data.DataLoader(
                data,
                batch_size=self.batch_size,
                shuffle=True,
                collate_fn=collator,
                num_workers=self.num_workers,
            ):
                self.zero_grad()

                if self.device.startswith("cuda"):
                    batch = move_to_cuda(batch)

                logits = self(batch)

                loss = F.nll_loss(
                    torch.log_softmax(logits, dim=1),
                    batch["labels"],
                    reduction="sum",
                )
                loss_accum += loss.item()
                n_samples_accum += len(batch["labels"])
                (loss / len(batch["labels"])).backward()
                optimizer.step()
                step += 1
                del loss

                if step % self.log_interval == 0:
                    loss_accum /= n_samples_accum
                    elapsed = time.perf_counter() - start_at
                    batch_per_sec = n_samples_accum / (
                        time.perf_counter() - start_at_accum
                    )
                    logger.info(
                        f"epoch:{epoch+1} step:{step} "
                        f"loss:{loss_accum:.2f} bps:{batch_per_sec:.2f} "
                        f"elapsed:{elapsed:.2f}"
                    )
                    loss_accum = 0
                    n_samples_accum = 0
                    start_at_accum = time.perf_counter()

                if (
                    self.save_every_step > -1
                    and self.saver
                    and step % self.save_every_step == 0
                ):
                    self.saver(self)

    def reset_module(self, **kwargs):
        pass

    def classifier_args(self):
        raise NotImplementedError

    def build_optimizer(self) -> torch.optim.Optimizer:
        unlinked_optimizer = self.kwargs.get(
            "optimizer",
            partial(torch.optim.AdamW),
        )
        optimizer: torch.optim.Optimizer = unlinked_optimizer(self.parameters())
        logger.info(f"Optimizer: {optimizer}")
        return optimizer

    def zero_grad(self):
        for p in self.parameters():
            p.grad = None

    def predict(self, X):
        self.to(self.device)
        self.eval()

        data = self.convert_data(X)

        y = []
        collator = Collator(self.pad)
        data_loader = torch.utils.data.DataLoader(
            data,
            batch_size=self.batch_size,
            shuffle=False,
            collate_fn=collator,
            num_workers=self.num_workers,
        )
        for batch in data_loader:
            if self.device.startswith("cuda"):
                batch = move_to_cuda(batch)

            with torch.inference_mode():
                maxi = torch.argmax(self(batch), dim=1)
            y.extend(maxi.tolist())
        return np.array(y)



import logging

import numpy as np
import torch
from torch.nn import Module

from bunruija.classifiers.classifier import NeuralBaseClassifier
from bunruija.modules import StaticEmbedding

logger = logging.getLogger(__name__)


class LSTMClassifier(NeuralBaseClassifier):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.embedding_path: str | None = kwargs.get("static_embedding_path", None)

        self.dim_emb: int = kwargs.get("dim_emb", 256)
        self.dim_hid: int = kwargs.get("dim_hid", 512)
        self.dropout_prob: float = kwargs.get("dropout", 0.15)
        self.num_layers: int = kwargs.get("num_layers", 1)
        self.bidirectional: bool = kwargs.get("bidirectional", True)

        if self.embedding_path:
            self.static_embed = StaticEmbedding(self.embedding_path)
        else:
            self.static_embed = None

        self.dropout: Module = torch.nn.Dropout(self.dropout_prob)

        self.lstm: Module = torch.nn.LSTM(
            input_size=(
                self.dim_emb + self.static_embed.dim_emb
                if self.static_embed
                else self.dim_emb
            ),
            hidden_size=self.dim_hid,
            num_layers=self.num_layers,
            bidirectional=self.bidirectional,
            dropout=self.dropout_prob if self.num_layers > 1 else 0.0,
        )

    def init_layer(self, data):
        max_input_idx = 0
        for data_i in data:
            max_input_idx = max(max_input_idx, np.max(data_i["inputs"]))

        self.embed = torch.nn.Embedding(
            max_input_idx + 1,
            self.dim_emb,
            padding_idx=0,
        )
        self.pad = 0

        self.fc1 = torch.nn.Linear(2 * self.dim_hid, self.dim_hid, bias=True)
        self.activation_fn = torch.nn.GELU()
        self.fc2 = torch.nn.Linear(self.dim_hid, len(self.labels), bias=True)

    def forward(self, batch):
        x = batch["inputs"]
        bsz = x.size(0)
        lengths = (x != self.pad).sum(dim=1)

        x = self.embed(x)
        if self.static_embed is not None:
            x_static = self.static_embed(batch)
            x_static = x_static.to(x.device)
            x = torch.cat([x, x_static], dim=2)

        x = self.dropout(x)

        packed = torch.nn.utils.rnn.pack_padded_sequence(
            x, lengths.cpu(), batch_first=True, enforce_sorted=False
        )

        _, (ht, _) = self.lstm(packed)

        ht = ht.view(2, self.num_layers, bsz, self.dim_hid)
        x = torch.cat(
            [ht[0, self.num_layers - 1, :, :], ht[1, self.num_layers - 1, :, :]], dim=1
        )
        x = self.dropout(x)
        x = self.fc1(x)
        x = self.activation_fn(x)
        x = self.fc2(x)
        return x



import argparse
from pathlib import Path


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("-y", "--yaml", help="Setting file")
    return parser


def get_default_preprocessing_parser():
    parser = get_parser()
    return parser


def get_default_train_parser():
    parser = get_parser()
    return parser


def get_default_prediction_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=Path, required=True)
    return parser


def get_default_evaluation_parser():
    parser = get_parser()
    parser.add_argument("--verbose", action="store_true", help="Verbose mode")
    parser.add_argument(
        "--no-evaluate-time",
        action="store_true",
        help="Disable evaluation of prediction time",
    )
    return parser


def get_default_gen_yaml_parser():
    parser = get_parser()
    parser.add_argument(
        "--model",
        default="sklearn.svm.SVC",
    )
    return parser



import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import PackedSequence


class QRNNLayer(torch.nn.Module):
    def __init__(self, input_size, output_size, window_size=2, bidirectional=True):
        super().__init__()

        self.num_gates = 3
        self.window_size = window_size
        self.input_size = input_size
        self.output_size = output_size
        self.bidirectional = bidirectional

        if self.bidirectional:
            self.fc = torch.nn.Linear(
                self.window_size * input_size, 2 * output_size * self.num_gates
            )
        else:
            self.fc = torch.nn.Linear(
                self.window_size * input_size, output_size * self.num_gates
            )

    def stride(self, x):
        """Create window_size sequences of embeddings.
        For the begining of sequence, insert padding embeddings to the left.
        For instance, sequence of window_size 3 will create the following output:

        Input: [1, 2, 3]
        Output: [[0, 0, 1]
                 [0, 1, 2]
                 [1, 2, 3]]
        """
        # Pad the left of bos
        pad_length = max(self.window_size - 1, 1)
        x = F.pad(x, (0, 0, pad_length, 0), value=0)

        bsz = x.size(0)
        seq_len = x.size(1)
        size = (
            bsz,
            seq_len - pad_length,
            self.input_size * self.window_size,
        )
        stride = x.stride()
        x = torch.as_strided(
            x,
            size,
            stride,
        )
        return x

    def forward(self, x):
        if isinstance(x, PackedSequence):
            return self.forward_packed_sequence(x)
        elif isinstance(x, torch.Tensor):
            return self.forward_tensor(x)

    def forward_packed_sequence(self, x):
        print(x)

    #         batch_sizes = x.batch_sizes
    #         num_steps = batch_sizes[0]
    #         print(batch_sizes)
    #         input_offset = 0
    #         pre_compute_input = self.fc(x.data)
    #         for i in range(num_steps):
    #             batch_size = batch_sizes[i]
    #             step_input = pre_compute_input.narrow(0, input_offset, batch_size)
    #             input_offset += batch_size
    #             print(i, batch_size, step_input.size())

    def forward_tensor(self, x: torch.Tensor):
        """
        Args:
            x (torch.Tensor): (bsz, seq_len, input_size)
        """
        bsz = x.size(0)
        seq_len = x.size(1)

        x = self.stride(x)
        x = x.view(bsz, seq_len, -1)
        x = self.fc(x)
        z, f, o = x.chunk(self.num_gates, dim=2)

        z = torch.tanh(z)
        f = torch.sigmoid(f)
        seq_len = z.size(1)

        c = torch.zeros_like(z)

        if self.bidirectional:
            c = c.view(bsz, seq_len, 2, self.output_size)
            f = f.view(bsz, seq_len, 2, self.output_size)
            z = z.view(bsz, seq_len, 2, self.output_size)
            # Forward QRNN
            direction = 0
            for t in range(seq_len):
                f_t = f[:, t, direction]
                z_t = z[:, t, direction]
                if t == 0:
                    c[:, t, direction] = (1 - f_t) * z_t
                else:
                    c_prev = c[:, t - 1, direction].clone()
                    c[:, t, direction] = f_t * c_prev + (1 - f_t) * z_t
            # backward QRNN
            direction = 1
            for t in range(seq_len - 1, -1, -1):
                f_t = f[:, t, direction]
                z_t = z[:, t, direction]
                if t == seq_len - 1:
                    c[:, t, direction] = (1 - f_t) * z_t
                else:
                    c_prev = c[:, t + 1, direction].clone()
                    c[:, t, direction] = f_t * c_prev + (1 - f_t) * z_t
            c = c.view(bsz, seq_len, 2 * self.output_size)
        else:
            for t in range(seq_len):
                if t == 0:
                    c[:, t] = (1 - f[:, t]) * z[:, t]
                else:
                    c[:, t] = f[:, t] * c[:, t - 1].clone() + (1 - f[:, t]) * z[:, t]

        h = torch.sigmoid(o) * c
        return h



from .lstm import LSTMClassifier
from .prado import PRADO
from .qrnn import QRNN
from .transformer import TransformerClassifier

__all__ = ["LSTMClassifier", "TransformerClassifier", "PRADO", "QRNN"]



import importlib
import logging
import os
import sys

import ruamel.yaml  # type: ignore

from . import options
from .classifiers import TransformerClassifier
from .classifiers.classifier import NeuralBaseClassifier
from .feature_extraction import SequenceVectorizer
from .tokenizers import MeCabTokenizer

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
    stream=sys.stdout,
)
logger = logging.getLogger("bunruija.gen_yaml")


def infer_vectorizer(model_cls):
    vectorizer_dict = {
        "type": None,
        "args": {
            "tokenizer": {
                "type": MeCabTokenizer.module_name(),
                "args": {"lemmatize": True},
            }
        },
    }

    if issubclass(model_cls, NeuralBaseClassifier):
        vectorizer_dict["type"] = SequenceVectorizer.module_name()
    else:
        vectorizer_dict["type"] = "sklearn.feature_extraction.text.TfidfVectorizer"
        vectorizer_dict["args"]["ngram_range"] = (1, 3)

    if model_cls == TransformerClassifier:
        vectorizer_dict["args"]["tokenizer"]["type"] = "transformers.AutoTokenizer"
        vectorizer_dict["args"]["tokenizer"]["args"] = {
            "pretrained_model_name_or_path": "cl-tohoku/bert-base-japanese"
        }

    return vectorizer_dict


def infer_model_args(model_cls):
    args = {}
    if model_cls == TransformerClassifier:
        args["pretrained_model_name_or_path"] = "cl-tohoku/bert-base-japanese"
    return args


def validate_model(model_name: str):
    module_elems: list[str] = model_name.split(".")
    module_name: str = ".".join(module_elems[:-1])
    cls_name: str = module_elems[-1]
    module = importlib.import_module(module_name)
    cls = getattr(module, cls_name, None)
    if cls is None:
        raise ValueError(model_name)
    return cls


def main(args):
    parser = options.get_default_gen_yaml_parser()
    args = parser.parse_args(args)

    model_cls = validate_model(args.model)

    setting = {
        "data": {
            "label_column": "label",
            "text_coliumn": "text",
            "args": {
                "path": "",
            },
        },
        "pipeline": [
            infer_vectorizer(model_cls),
            {
                "type": args.model,
                "args": infer_model_args(model_cls),
            },
        ],
    }

    if os.path.exists(args.yaml):
        raise FileExistsError(args.yaml)

    yaml = ruamel.yaml.YAML()
    with open(args.yaml, "w") as f:
        yaml.dump(setting, f)
    yaml.dump(setting, sys.stdout)


def cli_main():
    main(sys.argv[1:])


if __name__ == "__main__":
    cli_main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bunruija',
 'bunruija.classifiers',
 'bunruija.classifiers.prado',
 'bunruija.classifiers.qrnn',
 'bunruija.data',
 'bunruija.feature_extraction',
 'bunruija.filters',
 'bunruija.modules',
 'bunruija.tokenizers']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=3.0.8,<4.0.0',
 'datasets>=2.16.0,<3.0.0',
 'fugashi>=1.1.1,<2.0.0',
 'hydra-core>=1.2.0,<2.0.0',
 'ipadic>=1.0.0,<2.0.0',
 'lightgbm>=4.2.0,<5.0.0',
 'loguru>=0.7.2,<0.8.0',
 'mmh3>=3.0.0,<4.0.0',
 'ruamel-yaml>=0.18.5,<0.19.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'scipy>=1.8.0,<2.0.0',
 'torch>=2.1.2,<3.0.0',
 'transformers>=4.38.1,<5.0.0',
 'unidic-lite>=1.0.8,<2.0.0']

entry_points = \
{'console_scripts': ['bunruija-evaluate = bunruija.evaluate:cli_main',
                     'bunruija-gen-yaml = bunruija.gen_yaml:cli_main',
                     'bunruija-predict = bunruija.predict:cli_main',
                     'bunruija-train = bunruija.train:cli_main']}

setup_kwargs = {
    'name': 'bunruija',
    'version': '0.2.0',
    'description': 'A text classification toolkit',
    'long_description': '# Bunruija\n[![PyPI version](https://badge.fury.io/py/bunruija.svg)](https://badge.fury.io/py/bunruija)\n\nBunruija is a text classification toolkit.\nBunruija aims at enabling pre-processing, training and evaluation of text classification models with **minimum coding effort**.\nBunruija is mainly focusing on Japanese though it is also applicable to other languages.\n\nSee `example` for understanding how bunruija is easy to use.\n\n## Features\n- **Minimum requirements of coding**: bunruija enables users to train and evaluate their models through command lines. Because all experimental settings are stored in a yaml file, users do not have to write codes.\n- **Easy to compare neural-based model with non-neural-based model**: because bunruija supports models based on scikit-learn and PyTorch in the same framework, users can easily compare classification accuracies and prediction times of neural- and non-neural-based models.\n- **Easy to reproduce the training of a model**: because all hyperparameters of a model are stored in a yaml file, it is easy to reproduce the model.\n\n## Install\n```\npip install bunruija\n```\n\n## Example configs\nExample of `sklearn.svm.SVC`\n\n```yaml\ndata:\n  label_column: category\n  text_column: title\n  args:\n    path: data/jsonl\n\noutput_dir: models/svm-model\n\npipeline:\n  - type: sklearn.feature_extraction.text.TfidfVectorizer\n    args:\n      tokenizer:\n        type: bunruija.tokenizers.mecab_tokenizer.MeCabTokenizer\n        args:\n          lemmatize: true\n          exclude_pos:\n            - 助詞\n            - 助動詞\n      max_features: 10000\n      min_df: 3\n      ngram_range:\n        - 1\n        - 3\n  - type: sklearn.svm.SVC\n    args:\n      verbose: false\n      C: 10.\n```\n\nExample of BERT\n\n```yaml\ndata:\n  label_column: category\n  text_column: title\n  args:\n    path: data/jsonl\n\noutput_dir: models/transformer-model\n\npipeline:\n  - type: bunruija.feature_extraction.sequence.SequenceVectorizer\n    args:\n      tokenizer:\n        type: transformers.AutoTokenizer\n        args:\n          pretrained_model_name_or_path: cl-tohoku/bert-base-japanese\n  - type: bunruija.classifiers.transformer.TransformerClassifier\n    args:\n      device: cpu\n      pretrained_model_name_or_path: cl-tohoku/bert-base-japanese\n      optimizer:\n        type: torch.optim.AdamW\n        args:\n          lr: 3e-5\n          weight_decay: 0.01\n          betas:\n            - 0.9\n            - 0.999\n      max_epochs: 3\n```\n\n## CLI\n```sh\n# Training a classifier\nbunruija-train -y config.yaml\n\n# Evaluating the trained classifier\nbunruija-evaluate -y config.yaml\n```\n\n## Config\n### data\nYou can set data-related settings in `data`.\n\n```yaml\ndata:\n  label_column: category\n  text_column: title\n  args:\n    # Use local data in `data/jsonl`. In this path is assumed to contain data files such as train.jsonl, validation.jsonl and test.jsonl\n    path: data/jsonl\n\n    # If you want to use data on Hugging Face Hub, use the following args instead.\n    # Data is from https://huggingface.co/datasets/shunk031/livedoor-news-corpus\n    # path: shunk031/livedoor-news-corpus\n    # random_state: 0\n    # shuffle: true\n\n```\n\ndata is loaded via [datasets.load_dataset](https://huggingface.co/docs/datasets/main/en/package_reference/loading_methods#datasets.load_dataset).\nSo, you can load local data as well as data on [Hugging Face Hub](https://huggingface.co/datasets).\nWhen loading data, `args` are passed to `load_dataset`.\n\n`label_column` and `text_column` are field names of label and text.\n\nFormat of `csv`:\n\n```csv\ncategory,sentence\nsports,I like sports!\n…\n```\n\nFormat of `json`:\n\n```json\n[{"category", "sports", "text": "I like sports!"}]\n```\n\nFormat of `jsonl`:\n\n```json\n{"category", "sports", "text": "I like suports!"}\n```\n\n### pipeline\nYou can set pipeline of your model in `pipeline` section.\nIt is a list of components that are used in your model.\n\nFor each component, `type` is a module path and `args` is arguments for the module.\nFor instance, when you set the first component as follows, [TfidfVectorizer](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html) is instanciated with given arguments, and then applied to data at first in your model.\n\n```yaml\n  - type: sklearn.feature_extraction.text.TfidfVectorizer\n    args:\n      tokenizer:\n        type: bunruija.tokenizers.mecab_tokenizer.MeCabTokenizer\n        args:\n          lemmatize: true\n          exclude_pos:\n            - 助詞\n            - 助動詞\n      max_features: 10000\n      min_df: 3\n      ngram_range:\n        - 1\n        - 3\n```\n\n## Prediction using the trained classifier in Python code\nAfter you trained a classification model, you can use that model for prediction as follows:\n```python\nfrom bunruija import Predictor\n\npredictor = Predictor.from_pretrained("output_dir")\nwhile True:\n    text = input("Input:")\n    label: list[str] = predictor([text], return_label_type="str")\n    print(label[0])\n```\n\n`output_dir` is a directory that is specified in `output_dir` in config.\n',
    'author': 'Takuya Makino',
    'author_email': 'takuyamakino15@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tma15',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from typing import Any, Callable

import numpy as np
import transformers  # type: ignore
from scipy.sparse import csr_matrix  # type: ignore
from sklearn.base import TransformerMixin  # type: ignore

from ..data import Dictionary
from ..tokenizers import MeCabTokenizer


class SequencePairVectorizer(TransformerMixin):
    def __init__(
        self,
        tokenizer: Callable[[str], list[str]] | None = None,
        max_features: int | None = None,
        dictionary: Dictionary = Dictionary(),
        **kwargs,
    ):
        super().__init__()

        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.vocabulary_ = dictionary.index_to_element
        self.max_features = max_features

    @classmethod
    def module_name(cls) -> str:
        module_name = ".".join([cls.__module__, cls.__name__])
        return module_name

    def __repr__(self) -> str:
        args = []
        if self.tokenizer:
            args.append(f"tokenizer={self.tokenizer}")
        if self.max_features:
            args.append(f"max_features={self.max_features}")
        out = f'{self.__class__.__name__}({", ".join(args)})'
        return out

    def build_tokenizer(self) -> Callable[[str], list[str]]:
        if self.tokenizer is not None:
            return self.tokenizer

        self.tokenizer = MeCabTokenizer()
        return self.tokenizer

    def set_params(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def get_params(self, deep=True) -> dict[str, Any]:
        return {
            "tokenizer": self.tokenizer,
            "max_features": self.max_features,
            "dictionary": self.dictionary,
        }

    def fit(self, raw_documents: list[list[str]], y=None) -> "SequencePairVectorizer":
        tokenizer = self.build_tokenizer()

        for row_id, document in enumerate(raw_documents):
            elements = tokenizer(document[0])
            for element in elements:
                self.dictionary.add(element)

        if self.max_features is not None:
            filtered_dict = Dictionary()
            for k, v in sorted(
                zip(self.dictionary.elements, self.dictionary.count),
                key=lambda x: x[1],
                reverse=True,
            )[: self.max_features]:
                filtered_dict.add(k, v)

            self.dictionary = filtered_dict
        return self

    def transform(
        self,
        raw_documents: list[list[str]],
    ) -> csr_matrix | tuple[csr_matrix, list[str]]:
        data = []
        row = []
        col = []
        max_col = 0

        tokenizer = self.build_tokenizer()
        for row_id, document in enumerate(raw_documents):
            if isinstance(tokenizer, transformers.PreTrainedTokenizerBase):
                elements = tokenizer(
                    text=document[0], text_pair=document[1], truncation=True
                )
            else:
                elements_1 = tokenizer(document[0])
                elements_2 = tokenizer(document[1])
                elements = elements_1 + elements_2

            if isinstance(elements, transformers.tokenization_utils_base.BatchEncoding):
                input_ids = elements["input_ids"]
                max_col = max(max_col, len(input_ids))

                for i, index in enumerate(input_ids):
                    data.append(index)
                    row.append(row_id)
                    col.append(i)

            else:
                max_col = max(max_col, len(elements))

                for i, element in enumerate(elements):
                    if element in self.dictionary:
                        index = self.dictionary.get_index(element)
                        data.append(index)
                        row.append(row_id)
                        col.append(i)

        data = np.array(data)
        row = np.array(row)
        col = np.array(col)

        s = csr_matrix((data, (row, col)), shape=(len(raw_documents), max_col))
        return s

    def __call__(self, text):
        tokenizer = self.build_tokenizer()
        ret = tokenizer(text)
        return ret



from fugashi import Tagger  # type: ignore

from ..filters import PosFilter
from .tokenizer import BaseTokenizer


class MeCabTokenizer(BaseTokenizer):
    def __init__(self, **kwargs):
        super().__init__(name="mecab")
        self.lemmatize = kwargs.get("lemmatize", False)
        exclude_pos = kwargs.get("exclude_pos", [])
        if len(exclude_pos) > 0:
            self.filters = [PosFilter(exclude_pos)]
        else:
            self.filters = None

        self.dict_path = kwargs.get("dict_path", None)
        if self.dict_path:
            self._mecab = Tagger(f"-d {self.dict_path}")
        else:
            self._mecab = Tagger()

    def __getstate__(self):
        return {
            "lemmatize": self.lemmatize,
            "filters": self.filters,
            "dict_path": self.dict_path,
        }

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)

    def __getnewargs__(self):
        return ()

    def __reduce_ex__(self, proto):
        func = MeCabTokenizer
        args = self.__getnewargs__()
        state = self.__getstate__()
        listitems = None
        dictitems = None
        rv = (func, args, state, listitems, dictitems)
        return rv

    def __repr__(self) -> str:
        args = []
        args.append(f"lemmatize={self.lemmatize}")
        if self.dict_path:
            args.append(f"dict_path={self.dict_path}")
        if self.filters:
            args.append(f"filters={self.filters}")
        out = f'{self.__class__.__name__}({", ".join(args)})'
        return out

    def __call__(self, text: str) -> list[str]:
        ret: list[str] = []
        for word in self._mecab(text):
            if self.filters and any(
                [f(word.surface, word.feature) for f in self.filters]
            ):
                continue
            else:
                if self.lemmatize:
                    if word.feature.lemma is None:
                        ret.append(word.surface)
                    else:
                        ret.append(word.feature.lemma)
                else:
                    ret.append(word.surface)
        return ret



from collections import UserDict

import datasets
from datasets import (
    Dataset,
    DatasetDict,
    IterableDataset,
    IterableDatasetDict,
    load_dataset,
)


def load_data(
    dataset_args: UserDict,
    split: datasets.Split,
    label_column: str = "label",
    text_column: str | list[str] = "text",
) -> tuple[list[str], list[str] | list[list[str]]]:
    dataset: Dataset | DatasetDict | IterableDataset | IterableDatasetDict = (
        load_dataset(split=split, **dataset_args)
    )
    assert isinstance(dataset, Dataset)

    labels: list[str] = []
    texts: list[str] | list[list[str]]
    texts = []  # type: ignore

    for idx, sample in enumerate(dataset):
        label: str

        # If feature of label has names attribute, convert label to actual label strings
        if hasattr(dataset.features[label_column], "names"):
            label = dataset.features[label_column].names[sample[label_column]]
        else:
            label = sample[label_column]

        labels.append(label)

        if isinstance(text_column, str):
            input_example = sample[text_column]
            texts.append(input_example)
        elif isinstance(text_column, list):
            if len(text_column) != 2:
                raise ValueError(f"{len(text_column)=}")

            input_example = [sample[text_column[0]], sample[text_column[1]]]
            texts.append(input_example)
    return labels, texts



import pickle
from pathlib import Path

import numpy as np
from sklearn.preprocessing import LabelEncoder  # type: ignore


class Predictor:
    """Predicts labels"""

    def __init__(self, model_dir: str | Path):
        if isinstance(model_dir, str):
            model_dir = Path(model_dir)

        model_path: Path = model_dir / "model.bunruija"
        with open(model_path, "rb") as f:
            model_data: dict = pickle.load(f)
        self.model = model_data["pipeline"]
        self.label_encoder: LabelEncoder = model_data["label_encoder"]

    @classmethod
    def from_pretrained(cls, model_path: str | Path):
        return cls(model_path)

    def __call__(
        self,
        text: list[str],
        return_label_type: str = "id",
    ) -> np.ndarray | list[str]:
        y: np.ndarray = self.model.predict(text)

        if return_label_type == "str":
            label = self.label_encoder.inverse_transform(y)
        elif return_label_type == "id":
            label = y
        else:
            raise ValueError(return_label_type)
        return label



from .dataclass import BunruijaConfig
from .evaluator import Evaluator
from .feature_extraction.sequence import SequenceVectorizer
from .pipeline_builder import PipelineBuilder
from .predictor import Predictor
from .saver import Saver
from .trainer import Trainer

__all__ = [
    "BunruijaConfig",
    "Evaluator",
    "Predictor",
    "PipelineBuilder",
    "Saver",
    "SequenceVectorizer",
    "Trainer",
]



import importlib
from functools import partial
from logging import getLogger

import torch
from sklearn.pipeline import Pipeline  # type: ignore
from transformers import AutoTokenizer, PreTrainedTokenizer

from .dataclass import BunruijaConfig, PipelineUnit
from .saver import Saver

logger = getLogger(__name__)


class PipelineBuilder:
    def __init__(self, config: BunruijaConfig):
        self.config = config

    def _load_tokenizer(self, tokenizer_config: dict):
        name = tokenizer_config["type"]
        module_elems: list[str] = name.split(".")
        module_name: str = ".".join(module_elems[:-1])
        cls_name: str = module_elems[-1]
        module = importlib.import_module(module_name)
        cls = getattr(module, cls_name)
        tokenizer_args = tokenizer_config.get("args", {})
        if cls is AutoTokenizer or issubclass(cls, PreTrainedTokenizer):
            tokenizer = cls.from_pretrained(**tokenizer_args)
        else:
            tokenizer = cls(**tokenizer_args)

        return tokenizer

    def _load_class(self, name: str):
        module_elems: list[str] = name.split(".")
        module_name: str = ".".join(module_elems[:-1])
        cls_name: str = module_elems[-1]
        module = importlib.import_module(module_name)
        cls = getattr(module, cls_name)
        return cls

    def _maybe_build_tokenizer(self, pipeline_unit: PipelineUnit):
        """If a pipeline_unit has a tokenizer as an argument, build the tokenizer"""
        if "tokenizer" in pipeline_unit.args:
            tokenizer = self._load_tokenizer(pipeline_unit.args["tokenizer"])
            pipeline_unit.args["tokenizer"] = tokenizer

    def _maybe_update_arg(self, pipeline_unit: PipelineUnit):
        self._maybe_build_tokenizer(pipeline_unit)

        def _update_arg_value(x):
            if isinstance(x, list):
                return [_update_arg_value(_) for _ in x]
            elif isinstance(x, dict):
                if "tokenizer" in x.get("args", {}):
                    tokenizer = self._load_tokenizer(x["args"]["tokenizer"])
                    x["args"]["tokenizer"] = tokenizer

                # If type of x starts with pipeline, args of x is assumed to be list,
                # which is basically the list of weak learners in an ensemble model.
                if x["type"].startswith("pipeline"):
                    estimator = self.build_estimator(
                        [PipelineUnit(**_) for _ in x.get("args", [])],
                        pipeline_idx=x["type"],
                    )
                else:
                    estimator = self.build_estimator(PipelineUnit(**x))[1]
                return estimator
            else:
                return x

        for key, value in pipeline_unit.args.items():
            pipeline_unit.args[key] = _update_arg_value(value)
            cls = self._load_class(pipeline_unit.type)

            # Because some argments such as ngram_range in TfidfVectorizer assume tuple
            # while YAML cannot interpret as tuple, convert list to tuple according to
            # their constraints.
            if (
                hasattr(cls, "_parameter_constraints")
                and key in cls._parameter_constraints
            ):
                constraints = cls._parameter_constraints[key]
                if tuple in constraints and isinstance(value, list):
                    pipeline_unit.args[key] = tuple(pipeline_unit.args[key])

    def build_estimator(
        self,
        pipeline_units: PipelineUnit | list[PipelineUnit],
        pipeline_idx="pipeline",
    ):
        if isinstance(pipeline_units, list):
            estimators = [self.build_estimator(u) for u in pipeline_units]
            estimator_type = pipeline_idx
            memory = self.config.output_dir / "cache"
            estimator = Pipeline(estimators, memory=str(memory))
        else:
            self._maybe_update_arg(pipeline_units)
            estimator_type = pipeline_units.type
            cls = self._load_class(pipeline_units.type)

            # parameters of a neural network are not given at this moment.
            # so, partially create an optimizer
            if issubclass(cls, torch.optim.Optimizer):
                estimator = partial(cls, **pipeline_units.args)
            else:
                estimator = cls(**pipeline_units.args)

        # Because Pipeline of scikit-learn requires the tuple of name and estimator,
        # this functions returns them
        return estimator_type, estimator

    def build(self):
        setting = self.config.pipeline
        self.saver = Saver(self.config)

        model = self.build_estimator(setting)[1]
        logger.info(model)
        return model



import numpy as np
import torch
from transformers import AutoModel  # type: ignore
from transformers import AutoTokenizer

from .classifier import NeuralBaseClassifier


class TransformerClassifier(NeuralBaseClassifier):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        pretrained_model_name_or_path = kwargs.pop(
            "pretrained_model_name_or_path", None
        )
        self.model = AutoModel.from_pretrained(pretrained_model_name_or_path)
        self.dropout = torch.nn.Dropout(kwargs.get("dropout", 0.1))

        tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path)
        self.pad = tokenizer.pad_token_id

    def init_layer(self, data):
        y = []
        for data_i in data:
            y.append(data_i["label"])

        num_classes = np.unique(y)
        self.out = torch.nn.Linear(
            self.model.config.hidden_size, len(num_classes), bias=True
        )

    def forward(self, batch):
        input_ids = batch["inputs"]
        attention_mask = batch["attention_mask"]
        x = self.model(input_ids=input_ids, attention_mask=attention_mask)
        pooled_output = x[1]
        pooled_output = self.dropout(pooled_output)
        logits = self.out(pooled_output)
        return logits



import sys

from . import Predictor, options


def main(args):
    parser = options.get_default_prediction_parser()
    args = parser.parse_args(args)

    predictor = Predictor.from_pretrained(args.model)
    while True:
        text = input("Input:")
        label: list[str] = predictor([text], return_label_type="str")
        print(label[0])


def cli_main():
    main(sys.argv[1:])


if __name__ == "__main__":
    cli_main()
